@@ -1,6 +1,3 @@
-# -*- coding: UTF-8 -*-
-
-
 """
 emoji.core
 ~~~~~~~~~~
@@ -9,28 +6,59 @@ Core components for emoji.
 
 """
 
-import sys
-import unicodedata
 import re
+import unicodedata
+from typing import Iterator
 
 from emoji import unicode_codes
+from emoji.tokenizer import Token, EmojiMatch, EmojiMatchZWJ, EmojiMatchZWJNonRGI, tokenize, filter_tokens
 
 __all__ = [
-    'emojize', 'demojize',
+    'emojize', 'demojize', 'analyze', 'config',
     'emoji_list', 'distinct_emoji_list', 'emoji_count',
     'replace_emoji', 'is_emoji', 'version',
+    'Token', 'EmojiMatch', 'EmojiMatchZWJ', 'EmojiMatchZWJNonRGI',
 ]
 
-_SEARCH_TREE = None
 _DEFAULT_DELIMITER = ':'
-_EMOJI_NAME_PATTERN = u'\\w\\-&.’”“()!#*+?–,/«»\u0300\u0301\u0302\u0303\u0308\u030a\u0327\u064b\u064e\u064f\u0650\u0653\u0654\u3099\u30fb\u309a'
-_PY2 = sys.version_info[0] == 2
+_EMOJI_NAME_PATTERN = '\\w\\-&.’”“()!#*+?–,/«»\u0300\u0301\u0302\u0303\u0308\u030a\u0327\u064b\u064e\u064f\u0650\u0653\u0654\u3099\u30fb\u309a'
 
 
-def _normalize(form, s):
-    if _PY2:
-        s = unicode(s)
-    return unicodedata.normalize(form, s)
+class config():
+    """Module-wide configuration"""
+
+    demojize_keep_zwj = True
+    """Change the behavior of :func:`emoji.demojize()` regarding
+    zero-width-joiners (ZWJ/``\\u200D``) in emoji that are not
+    "recommended for general interchange" (non-RGI).
+    It has no effect on RGI emoji.
+
+    For example this family emoji with different skin tones "👨‍👩🏿‍👧🏻‍👦🏾" contains four
+    person emoji that are joined together by three ZWJ characters:
+    ``👨\\u200D👩🏿\\u200D👧🏻\\u200D👦🏾``
+
+    If ``True``, the zero-width-joiners will be kept and :func:`emoji.emojize()` can
+    reverse the :func:`emoji.demojize()` operation:
+    ``emoji.emojize(emoji.demojize(s)) == s``
+
+    The example emoji would be converted to
+    ``:man:\\u200d:woman_dark_skin_tone:\\u200d:girl_light_skin_tone:\\u200d:boy_medium-dark_skin_tone:``
+
+    If ``False``, the zero-width-joiners will be removed and :func:`emoji.emojize()`
+    can only reverse the individual emoji: ``emoji.emojize(emoji.demojize(s)) != s``
+
+    The example emoji would be converted to
+    ``:man::woman_dark_skin_tone::girl_light_skin_tone::boy_medium-dark_skin_tone:``
+    """
+
+    replace_emoji_keep_zwj = False
+    """Change the behavior of :func:`emoji.replace_emoji()` regarding
+    zero-width-joiners (ZWJ/``\\u200D``) in emoji that are not
+    "recommended for general interchange" (non-RGI).
+    It has no effect on RGI emoji.
+
+    See :attr:`config.demojize_keep_zwj` for more information.
+    """
 
 
 def emojize(
@@ -67,16 +95,16 @@ def emojize(
     :param handle_version: (optional) Replace the emoji above ``version``
         instead of ignoring it. handle_version can be either a string or a
         callable; If it is a callable, it's passed the Unicode emoji and the
-        data dict from emoji.EMOJI_DATA and must return a replacement string
+        data dict from :data:`EMOJI_DATA` and must return a replacement string
         to be used::
 
-            handle_version(u'\\U0001F6EB', {
+            handle_version('\\U0001F6EB', {
                 'en' : ':airplane_departure:',
                 'status' : fully_qualified,
                 'E' : 1,
-                'alias' : [u':flight_departure:'],
-                'de': u':abflug:',
-                'es': u':avión_despegando:',
+                'alias' : [':flight_departure:'],
+                'de': ':abflug:',
+                'es': ':avión_despegando:',
                 ...
             })
 
@@ -89,12 +117,15 @@ def emojize(
     else:
         language_pack = unicode_codes.get_emoji_unicode_dict(language)
 
-    pattern = re.compile(u'(%s[%s]+%s)' %
-                         (re.escape(delimiters[0]), _EMOJI_NAME_PATTERN, re.escape(delimiters[1])), flags=re.UNICODE)
+    pattern = re.compile('(%s[%s]+%s)' %
+                         (re.escape(delimiters[0]), _EMOJI_NAME_PATTERN, re.escape(delimiters[1])))
 
     def replace(match):
         name = match.group(1)[len(delimiters[0]):-len(delimiters[1])]
-        emj = language_pack.get(_DEFAULT_DELIMITER + _normalize('NFKC', name) + _DEFAULT_DELIMITER)
+        emj = language_pack.get(
+            _DEFAULT_DELIMITER +
+            unicodedata.normalize('NFKC', name) +
+            _DEFAULT_DELIMITER)
         if emj is None:
             return match.group(1)
 
@@ -113,13 +144,13 @@ def emojize(
         if variant is None or 'variant' not in unicode_codes.EMOJI_DATA[emj]:
             return emj
 
-        if emj[-1] == u'\uFE0E' or emj[-1] == u'\uFE0F':
+        if emj[-1] == '\uFE0E' or emj[-1] == '\uFE0F':
             # Remove an existing variant
             emj = emj[0:-1]
         if variant == "text_type":
-            return emj + u'\uFE0E'
+            return emj + '\uFE0E'
         elif variant == "emoji_type":
-            return emj + u'\uFE0F'
+            return emj + '\uFE0F'
         else:
             raise ValueError(
                 "Parameter 'variant' must be either None, 'text_type' or 'emoji_type'")
@@ -127,6 +158,23 @@ def emojize(
     return pattern.sub(replace, string)
 
 
+def analyze(string: str, non_emoji: bool = False, join_emoji: bool = True) -> Iterator[Token]:
+    """
+    Find unicode emoji in a string. Yield each emoji as a named tuple
+    :class:`Token` ``(chars, EmojiMatch)`` or `:class:`Token` ``(chars, EmojiMatchZWJNonRGI)``.
+    If ``non_emoji`` is True, also yield all other characters as
+    :class:`Token` ``(char, char)`` .
+
+    :param string: String to analyze
+    :param non_emoji: If True also yield all non-emoji characters as Token(char, char)
+    :param join_emoji: If True, multiple EmojiMatch are merged into a single
+        EmojiMatchZWJNonRGI if they are separated only by a ZWJ.
+    """
+
+    return filter_tokens(
+        tokenize(string, keep_zwj=True), emoji_only=not non_emoji, join_emoji=join_emoji)
+
+
 def demojize(
         string,
         delimiters=(_DEFAULT_DELIMITER, _DEFAULT_DELIMITER),
@@ -139,9 +187,9 @@ def demojize(
         >>> import emoji
         >>> print(emoji.emojize("Python is fun :thumbs_up:"))
         Python is fun 👍
-        >>> print(emoji.demojize(u"Python is fun 👍"))
+        >>> print(emoji.demojize("Python is fun 👍"))
         Python is fun :thumbs_up:
-        >>> print(emoji.demojize(u"Unicode is tricky 😯", delimiters=("__", "__")))
+        >>> print(emoji.demojize(icode is tricky 😯", delimiters=("__", "__")))
         Unicode is tricky __hushed_face__
 
     :param string: String contains Unicode characters. MUST BE UNICODE.
@@ -154,16 +202,16 @@ def demojize(
         instead of removing it. handle_version can be either a string or a
         callable ``handle_version(emj: str, data: dict) -> str``; If it is
         a callable, it's passed the Unicode emoji and the data dict from
-        emoji.EMOJI_DATA and must return a replacement string  to be used.
+        :data:`EMOJI_DATA` and must return a replacement string  to be used.
         The passed data is in the form of::
 
-            handle_version(u'\\U0001F6EB', {
+            handle_version('\\U0001F6EB', {
                 'en' : ':airplane_departure:',
                 'status' : fully_qualified,
                 'E' : 1,
-                'alias' : [u':flight_departure:'],
-                'de': u':abflug:',
-                'es': u':avión_despegando:',
+                'alias' : [':flight_departure:'],
+                'de': ':abflug:',
+                'es': ':avión_despegando:',
                 ...
             })
 
@@ -175,78 +223,60 @@ def demojize(
     else:
         _use_aliases = False
 
-    tree = _get_search_tree()
-    result = []
-    i = 0
-    length = len(string)
-    while i < length:
-        consumed = False
-        char = string[i]
-        if char in tree:
-            j = i + 1
-            sub_tree = tree[char]
-            while j < length and string[j] in sub_tree:
-                sub_tree = sub_tree[string[j]]
-                j += 1
-            if 'data' in sub_tree:
-                emj_data = sub_tree['data']
-                code_points = string[i:j]
-                replace_str = None
-                if version is not None and emj_data['E'] > version:
-                    if callable(handle_version):
-                        emj_data = emj_data.copy()
-                        emj_data['match_start'] = i
-                        emj_data['match_end'] = j
-                        replace_str = handle_version(code_points, emj_data)
-                    elif handle_version is not None:
-                        replace_str = str(handle_version)
-                    else:
-                        replace_str = None
-                elif language in emj_data:
-                    if _use_aliases and 'alias' in emj_data:
-                        replace_str = delimiters[0] + emj_data['alias'][0][1:-1] + delimiters[1]
-                    else:
-                        replace_str = delimiters[0] + emj_data[language][1:-1] + delimiters[1]
-                else:
-                    # The emoji exists, but it is not translated, so we keep the emoji
-                    replace_str = code_points
+    def handle(emoji_match):
+        if version is not None and emoji_match.data['E'] > version:
+            if callable(handle_version):
+                return handle_version(emoji_match.emoji, emoji_match.data_copy())
+            elif handle_version is not None:
+                return handle_version
+            else:
+                return ''
+        elif language in emoji_match.data:
+            if _use_aliases and 'alias' in emoji_match.data:
+                return delimiters[0] + emoji_match.data['alias'][0][1:-1] + delimiters[1]
+            else:
+                return delimiters[0] + emoji_match.data[language][1:-1] + delimiters[1]
+        else:
+            # The emoji exists, but it is not translated, so we keep the emoji
+            return emoji_match.emoji
 
-                i = j - 1
-                consumed = True
-                if replace_str:
-                    result.append(replace_str)
-
-        if not consumed and char != u'\ufe0e' and char != u'\ufe0f':
-            result.append(char)
-        i += 1
-
-    return "".join(result)
+    matches = tokenize(string, keep_zwj=config.demojize_keep_zwj)
+    return "".join(str(handle(token.value)) if isinstance(
+        token.value, EmojiMatch) else token.value for token in matches)
 
 
-def replace_emoji(string, replace='',  version=-1):
+def replace_emoji(string, replace='', version=-1):
     """
     Replace Unicode emoji in a customizable string.
 
     :param string: String contains Unicode characters. MUST BE UNICODE.
     :param replace: (optional) replace can be either a string or a callable;
         If it is a callable, it's passed the Unicode emoji and the data dict from
-        emoji.EMOJI_DATA and must return a replacement string to be used.
+        :data:`EMOJI_DATA` and must return a replacement string to be used.
         replace(str, dict) -> str
     :param version: (optional) Max version. If set to an Emoji Version,
         only emoji above this version will be replaced.
     """
 
-    if version > -1:
-        def f(emj, emj_data):
-            if emj_data['E'] <= version:
-                return emj  # Do not replace emj
-            if callable(replace):
-                return replace(emj, emj_data)
-            return str(replace)
+    def handle(emoji_match):
+        if version > -1:
+            if emoji_match.data['E'] > version:
+                if callable(replace):
+                    return replace(emoji_match.emoji, emoji_match.data_copy())
+                else:
+                    return str(replace)
+        elif callable(replace):
+            return replace(emoji_match.emoji, emoji_match.data_copy())
+        elif replace is not None:
+            return replace
+        return emoji_match.emoji
 
-        return demojize(string, language='en', version=-1, handle_version=f)
-    else:
-        return demojize(string, language='en', version=-1, handle_version=replace)
+    matches = tokenize(string, keep_zwj=config.replace_emoji_keep_zwj)
+    if config.replace_emoji_keep_zwj:
+        matches = filter_tokens(
+            matches, emoji_only=False, join_emoji=True)
+    return "".join(str(handle(m.value)) if isinstance(
+        m.value, EmojiMatch) else m.value for m in matches)
 
 
 def emoji_list(string):
@@ -255,17 +285,12 @@ def emoji_list(string):
         >>> emoji.emoji_list("Hi, I am fine. 😁")
         [{'match_start': 15, 'match_end': 16, 'emoji': '😁'}]
     """
-    _entities = []
 
-    def f(emj, emj_data):
-        _entities.append({
-            'match_start': emj_data['match_start'],
-            'match_end': emj_data['match_end'],
-            'emoji': emj,
-        })
-
-    demojize(string, language='en', version=-1, handle_version=f)
-    return _entities
+    return [{
+        'match_start': m.value.start,
+        'match_end': m.value.end,
+        'emoji': m.value.emoji,
+    } for m in tokenize(string, keep_zwj=False) if isinstance(m.value, EmojiMatch)]
 
 
 def distinct_emoji_list(string):
@@ -288,7 +313,10 @@ def emoji_count(string, unique=False):
 
 
 def is_emoji(string):
-    """Returns True if the string is an emoji."""
+    """
+    Returns True if the string is an emoji and it is "recommended for
+    general interchange" by Unicode.org.
+    """
     return string in unicode_codes.EMOJI_DATA
 
 
@@ -302,7 +330,7 @@ def version(string):
         >>> emoji.version(":butterfly:")
         3
 
-    :param string: An emoji or a text containig an emoji
+    :param string: An emoji or a text containing an emoji
     :raises ValueError: if ``string`` does not contain an emoji
     """
     # Try dictionary lookup
@@ -333,66 +361,3 @@ def version(string):
             return version[0]
 
     raise ValueError("No emoji found in string")
-
-
-def _get_search_tree():
-    """
-    Generate a search tree for demojize().
-    Example of a search tree::
-
-        EMOJI_DATA =
-        {'a': {'en': ':Apple:'},
-        'b': {'en': ':Bus:'},
-        'ba': {'en': ':Bat:'},
-        'band': {'en': ':Beatles:'},
-        'bandit': {'en': ':Outlaw:'},
-        'bank': {'en': ':BankOfEngland:'},
-        'bb': {'en': ':BB-gun:'},
-        'c': {'en': ':Car:'}}
-
-        _SEARCH_TREE =
-        {'a': {'data': {'en': ':Apple:'}},
-        'b': {'a': {'data': {'en': ':Bat:'},
-                    'n': {'d': {'data': {'en': ':Beatles:'},
-                                'i': {'t': {'data': {'en': ':Outlaw:'}}}},
-                        'k': {'data': {'en': ':BankOfEngland:'}}}},
-            'b': {'data': {'en': ':BB-gun:'}},
-            'data': {'en': ':Bus:'}},
-        'c': {'data': {'en': ':Car:'}}}
-
-                   _SEARCH_TREE
-                 /     |        ⧵
-               /       |          ⧵
-            a          b             c
-            |        / |  ⧵          |
-            |       /  |    ⧵        |
-        :Apple:   ba  :Bus:  bb     :Car:
-                 /  ⧵         |
-                /    ⧵        |
-              :Bat:    ban     :BB-gun:
-                     /     ⧵
-                    /       ⧵
-                 band       bank
-                /   ⧵         |
-               /     ⧵        |
-            bandi :Beatles:  :BankOfEngland:
-               |
-            bandit
-               |
-           :Outlaw:
-
-
-    """
-    global _SEARCH_TREE
-    if _SEARCH_TREE is None:
-        _SEARCH_TREE = {}
-        for emj in unicode_codes.EMOJI_DATA:
-            sub_tree = _SEARCH_TREE
-            lastidx = len(emj) - 1
-            for i, char in enumerate(emj):
-                if char not in sub_tree:
-                    sub_tree[char] = {}
-                sub_tree = sub_tree[char]
-                if i == lastidx:
-                    sub_tree['data'] = unicode_codes.EMOJI_DATA[emj]
-    return _SEARCH_TREE
